-- File: stdio_head_h.vhd
-- Version: 2.0	 (March 13, 2002)
-- Source: http://bear.ces.cwru.edu/vhdl
-- Date: March 13, 2002 (Copyright)
-- Author: Francis G. Wolff   Email: fxw12@po.cwru.edu
-- Author: Michael J. Knieser Email: mjknieser@knieser.com
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 1, or (at your option)
-- any later version: http://www.gnu.org/licenses/gpl.html
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
use     std.textio.all;
library ieee;
use     ieee.std_logic_1164.all;
use     ieee.std_logic_textio.all;
library C;
use     C.strings_h.all;
use     C.regexp_h.all;
use     C.stdlib_h.all; --atoi

package stdio_h is
  shared    variable printf_buffer:  line;
  shared    variable fprintf_buffer: line;
  shared    variable fscanf_buffer:  line;
  shared    variable scanf_buffer:   line;

  function pf(a1: IN bit)               return string;
  function pf(a1: IN boolean)           return string;
  function pf(a1: IN character)         return string;
  function pf(a1: IN std_ulogic)        return string;
  function pf(a1: IN string)            return string;
  function pf(a1: IN integer)           return string;
  function pf(a1: IN bit_vector)        return string;
  function pf(a1: IN std_ulogic_vector) return string;
  function pf(a1: IN std_logic_vector)  return string;
  function pf(a1: IN time)              return string;

  procedure sbufprintf(sbuf:   IN boolean; fi:  INOUT integer; buf: INOUT line;
                       file stream: text; fmt: IN string;     s1: IN string);

  procedure sbufscanf_id( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT integer);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT line);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT character);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT string);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT integer);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT boolean);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT bit);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT bit_vector);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT std_logic);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT std_logic_vector);

  procedure printf(format:  IN	  string);

  procedure fprintf(buf:     INOUT line; --need for concurrent multi-file writes
                    file stream:  text;
                    format:  IN    string;
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure fprintf(file stream:  text; --limited to a single write file at a time
                    format:  IN    string;
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure sprintf(s: INOUT line; format: IN string; 
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure sprintf(s: INOUT string; format: IN string; 
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure fgetc(  c: OUT character; file stream: text; buf: INOUT line);
  procedure fgetc(  c: OUT character; file stream: text);
  procedure getc(   c: OUT character; file stream: text);
  procedure ungetc( c: IN  character; file stream: text; buf: INOUT line);
  procedure getchar(c: OUT character);
  procedure fgets(  s: OUT  string; n: IN integer; file stream: text; buf: INOUT line);
  procedure fgets(  s: OUT  string; n: IN integer; file stream: text);
  procedure gets(   s: OUT string);
  procedure fputc(  c: IN character; file stream: text; buf: INOUT line);
  procedure fputc(  c: IN character; file stream: text);
  procedure putc(   c: IN character; file stream: text);
  procedure putchar(c: IN character);
  procedure fputs(  s: IN string; file stream: text; buf: INOUT line);
  procedure fputs(  s: IN string; file stream: text);
  procedure puts(   s: IN string);

  procedure fclose(buf: INOUT line; file stream: text);
  procedure fclose(file stream: text);

