-- File: stdio_head_h.vhd
-- Version: 1.0	 (March 13, 2002)
-- Source: http://bear.ces.cwru.edu/vhdl
-- Date: March 13, 2002 (Copyright)
-- Author: Francis G. Wolff   Email: fxw12@po.cwru.edu
-- Author: Michael J. Knieser Email: mjknieser@knieser.com
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 1, or (at your option)
-- any later version: http://www.gnu.org/licenses/gpl.html
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
use     std.textio.all;
library ieee;
use     ieee.std_logic_1164.all;
use     ieee.std_logic_textio.all;
library C;
use     C.strings_h.all;
use     C.regexp_h.all;
use     C.stdlib_h.all; --atoi

package stdio_h is
  shared    variable printf_buffer:  line;
  shared    variable fprintf_buffer: line;
  shared    variable fscanf_buffer:  line;
  shared    variable scanf_buffer:   line;

  function pf(a1: IN bit)               return string;
  function pf(a1: IN boolean)           return string;
  function pf(a1: IN character)         return string;
  function pf(a1: IN std_ulogic)        return string;
  function pf(a1: IN string)            return string;
  function pf(a1: IN integer)           return string;
  function pf(a1: IN bit_vector)        return string;
  function pf(a1: IN std_ulogic_vector) return string;
  function pf(a1: IN std_logic_vector)  return string;
  function pf(a1: IN time)              return string;

  procedure sbufprintf(sbuf:   IN boolean; fi:  INOUT integer; buf: INOUT line;
                       file stream: text; fmt: IN string;     s1: IN string);

  procedure sbufscanf_id( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT integer);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT line);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT character);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT string);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT integer);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT boolean);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT bit);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT bit_vector);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT std_logic);

  procedure sbufscanf( sbuf:  IN boolean;  fi:  INOUT integer; buf: INOUT line;
                      file stream: text;  fmt: IN string;     s1:  INOUT std_logic_vector);

  procedure printf(format:  IN	  string);

  procedure fprintf(buf:     INOUT line; --need for concurrent multi-file writes
                    file stream:  text;
                    format:  IN    string;
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure fprintf(file stream:  text; --limited to a single write file at a time
                    format:  IN    string;
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure sprintf(s: INOUT line; format: IN string; 
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure sprintf(s: INOUT string; format: IN string; 
                    a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : IN string := " ";
                    a9,  a10, a11, a12, a13, a14, a15, a16: IN string := " ");

  procedure fgetc(  c: OUT character; file stream: text; buf: INOUT line);
  procedure fgetc(  c: OUT character; file stream: text);
  procedure getc(   c: OUT character; file stream: text);
  procedure ungetc( c: IN  character; file stream: text; buf: INOUT line);
  procedure getchar(c: OUT character);
  procedure fgets(  s: OUT  string; n: IN integer; file stream: text; buf: INOUT line);
  procedure fgets(  s: OUT  string; n: IN integer; file stream: text);
  procedure gets(   s: OUT string);
  procedure fputc(  c: IN character; file stream: text; buf: INOUT line);
  procedure fputc(  c: IN character; file stream: text);
  procedure putc(   c: IN character; file stream: text);
  procedure putchar(c: IN character);
  procedure fputs(  s: IN string; file stream: text; buf: INOUT line);
  procedure fputs(  s: IN string; file stream: text);
  procedure puts(   s: IN string);

  procedure fclose(buf: INOUT line; file stream: text);
  procedure fclose(file stream: text);

  procedure printf(format: IN string; a1: string);

  procedure printf(format: IN string; a1: integer);

  procedure printf(format: IN string; a1: std_logic);

  procedure printf(format: IN string; a1: std_logic_vector);

  procedure printf(format: IN string; a1: boolean);

  procedure printf(format: IN string; a1: bit);

  procedure printf(format: IN string; a1: bit_vector);

  procedure printf(format: IN string; a1: string; a2: string);

  procedure printf(format: IN string; a1: integer; a2: string);

  procedure printf(format: IN string; a1: std_logic; a2: string);

  procedure printf(format: IN string; a1: std_logic_vector; a2: string);

  procedure printf(format: IN string; a1: boolean; a2: string);

  procedure printf(format: IN string; a1: bit; a2: string);

  procedure printf(format: IN string; a1: bit_vector; a2: string);

  procedure printf(format: IN string; a1: string; a2: integer);

  procedure printf(format: IN string; a1: integer; a2: integer);

  procedure printf(format: IN string; a1: std_logic; a2: integer);

  procedure printf(format: IN string; a1: std_logic_vector; a2: integer);

  procedure printf(format: IN string; a1: boolean; a2: integer);

  procedure printf(format: IN string; a1: bit; a2: integer);

  procedure printf(format: IN string; a1: bit_vector; a2: integer);

  procedure printf(format: IN string; a1: string; a2: std_logic);

  procedure printf(format: IN string; a1: integer; a2: std_logic);

  procedure printf(format: IN string; a1: std_logic; a2: std_logic);

  procedure printf(format: IN string; a1: std_logic_vector; a2: std_logic);

  procedure printf(format: IN string; a1: boolean; a2: std_logic);

  procedure printf(format: IN string; a1: bit; a2: std_logic);

  procedure printf(format: IN string; a1: bit_vector; a2: std_logic);

  procedure printf(format: IN string; a1: string; a2: std_logic_vector);

  procedure printf(format: IN string; a1: integer; a2: std_logic_vector);

  procedure printf(format: IN string; a1: std_logic; a2: std_logic_vector);

  procedure printf(format: IN string; a1: std_logic_vector; a2: std_logic_vector);

  procedure printf(format: IN string; a1: boolean; a2: std_logic_vector);

  procedure printf(format: IN string; a1: bit; a2: std_logic_vector);

  procedure printf(format: IN string; a1: bit_vector; a2: std_logic_vector);

  procedure printf(format: IN string; a1: string; a2: boolean);

  procedure printf(format: IN string; a1: integer; a2: boolean);

  procedure printf(format: IN string; a1: std_logic; a2: boolean);

  procedure printf(format: IN string; a1: std_logic_vector; a2: boolean);

  procedure printf(format: IN string; a1: boolean; a2: boolean);

  procedure printf(format: IN string; a1: bit; a2: boolean);

  procedure printf(format: IN string; a1: bit_vector; a2: boolean);

  procedure printf(format: IN string; a1: string; a2: bit);

  procedure printf(format: IN string; a1: integer; a2: bit);

  procedure printf(format: IN string; a1: std_logic; a2: bit);

  procedure printf(format: IN string; a1: std_logic_vector; a2: bit);

  procedure printf(format: IN string; a1: boolean; a2: bit);

  procedure printf(format: IN string; a1: bit; a2: bit);

  procedure printf(format: IN string; a1: bit_vector; a2: bit);

  procedure printf(format: IN string; a1: string; a2: bit_vector);

  procedure printf(format: IN string; a1: integer; a2: bit_vector);

  procedure printf(format: IN string; a1: std_logic; a2: bit_vector);

  procedure printf(format: IN string; a1: std_logic_vector; a2: bit_vector);

  procedure printf(format: IN string; a1: boolean; a2: bit_vector);

  procedure printf(format: IN string; a1: bit; a2: bit_vector);

  procedure printf(format: IN string; a1: bit_vector; a2: bit_vector);

  procedure scanf(format: IN string; a1: INOUT string);

  procedure scanf(format: IN string; a1: INOUT integer);

  procedure scanf(format: IN string; a1: INOUT std_logic);

  procedure scanf(format: IN string; a1: INOUT std_logic_vector);

  procedure scanf(format: IN string; a1: INOUT boolean);

  procedure scanf(format: IN string; a1: INOUT bit);

  procedure scanf(format: IN string; a1: INOUT bit_vector);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT string);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT integer);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT std_logic);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT std_logic_vector);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT boolean);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT bit);

  procedure fscanf(file stream: text; format: IN string; a1: INOUT bit_vector);

  impure function sscanf(s: string; format: string) return integer;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT string);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT integer);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT std_logic);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT std_logic_vector);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT boolean);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT bit);

  procedure sscanf(s: IN string; format: IN string; a1: INOUT bit_vector);

end stdio_h;
 
-- File:    stdio_h_tail.vhd
-- Version: 1.0	 (March 13, 2002)
-- Source:  http://bear.ces.cwru.edu/vhdl
-- Date:    March 13, 2002 (Copyright)
-- Author:  Francis G. Wolff   Email: fxw12@po.cwru.edu
-- Author:  Michael J. Knieser Email: mjknieser@knieser.com
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 1, or (at your option)
-- any later version: http://www.gnu.org/licenses/gpl.html
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--
package body stdio_h is

  type     pf_bit_vector_type is    array(bit) of character;
  constant pf_bit_vector:           pf_bit_vector_type:= ('0', '1');

  type     pf_std_logic_vector_type is array(std_ulogic) of character;
  constant pf_std_logic_vector:        pf_std_logic_vector_type
             := ('U', 'X', '0', '1', 'Z', 'W', 'L', 'H', '-');

  type     pf_hex_type is array(0 to 15) of character;
  constant pf_hex: pf_hex_type
        := ('0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f');

  type     pf_hex2_type is array(0 to 15) of string(1 to 4);
  constant pf_hex2: pf_hex2_type
        := ("0000","1000","0100","1100","0010","1010","0110","1110",
            "0001","1001","0101","1101","0011","1011","0111","1111");

  function pf(a1: IN bit) return string is
    variable s: string(1 to 1);
  begin
    s(1):=pf_bit_vector(a1); return s;
  end pf;

  function pf(a1: IN boolean) return string is
    variable s: string(1 to 1);
  begin
    if a1 then s(1):='1'; else s(1):='0'; end if;
    return s;
  end pf;

  function pf(a1: IN character) return string is
    variable s: string(1 to 1);
  begin
    s(1):=a1; return s;
  end pf;

  function pf(a1: IN std_ulogic) return string is
    variable s: string(1 to 1);
  begin
    s(1):=pf_std_logic_vector(a1); return s;
  end pf;

  function pf(a1: IN string) return string is
  begin
    return a1;
  end pf;

  function pf2(s: IN string) return string is
    variable t: string(1 to s'length):=s;
    variable c: boolean:=true;
    variable i: integer:=1;
  begin --2's complement
    while i <= s'length loop
      if    s(i)='0' then
        if c then t(i):='0'; else t(i):='1'; c:=false; end if; 
      elsif s(i)='1' then
        if c then t(i):='1'; else t(i):='0'; end if; c:=false;
      elsif s(i)=NUL then
         exit;
      end if;
      i:=i+1;
    end loop;
    if i<=t'length AND not c then t(i):='1'; i:=i+1; end if;
    if i<=t'length           then t(i):=NUL; end if;
    return t;
  end pf2;

  function pf(a1: IN integer) return string is
    variable ax, j: integer:=0;
    variable r:     string(1 to 64);
  begin
    j:=0; ax:=a1;
    if a1<0 then ax:=-a1; else ax:=a1; end if;
    loop
      if j<r'length     then j:=j+1; else exit; end if;
      if (ax mod 2) = 0 then r(j):='0'; else r(j):='1'; end if;
      ax:=ax/2; if ax=0 then exit; end if;
    end loop;
    --underscore not dash (see std_logic)
    if j<r'length          then j:=j+1; r(j):=NUL; end if;
    if a1<0                then r:=pf2(r); else r(j):='0'; j:=j+1; r(j):=NUL; end if;
    return r;
  end pf;

  function pf(a1: IN bit_vector) return string is
    variable r: string(1 to a1'length);
    variable j: integer:=1;
  begin
    if a1'left<=a1'right then --little endian
      for i in a1'left to a1'right loop
        r(j) := pf_bit_vector(a1(i)); j:=j+1;
      end loop;
    else
      for i in a1'left downto a1'right loop
        r(j) := pf_bit_vector(a1(i)); j:=j+1;
      end loop;
    end if;
    return r;
  end pf;

  --type conversion --keep little endian format
  function pf(a1: IN std_ulogic_vector) return string is
    variable r: string(1 to a1'length);
    variable j: integer:=1;
  begin
    if a1'left<=a1'right then --little endian
      for i in a1'left to a1'right loop
        r(j) := pf_std_logic_vector(a1(i)); j:=j+1;
      end loop;
    else
      for i in a1'left downto a1'right loop
        r(j) := pf_std_logic_vector(a1(i)); j:=j+1;
      end loop;
    end if;
    return r;
  end pf;

  --type conversion --keep endian format
  function pf(a1: IN std_logic_vector) return string is
    variable r: string(1 to a1'length);
    variable j: integer:=1;
  begin
    if a1'left <= a1'right then --little endian
      for i in a1'left to a1'right loop
        r(j) := pf_std_logic_vector(a1(i)); j:=j+1;
      end loop;
    else
      for i in a1'left downto a1'right loop
        r(j) := pf_std_logic_vector(a1(i)); j:=j+1;
      end loop;
    end if;
    return r;
  end pf;

  function pf(a1: IN time) return string is
  begin
    return pf(integer(a1 / 1 ns));   
  end pf;

  function sf(s1: IN string) return integer is
    variable i, n, ap, ax: integer;
    variable d:            character;
  begin
    ax:=0; d:=' ';
    for i in 1 to s1'length loop
       if s1(i)=NUL then exit;                 end if;
       if s1(i)=' ' then exit;                 end if;
       if i=1       then ap:=1; else ap:=ap*2; end if;
       if s1(i)='1' then ax:=ax+ap;            end if;
       d:=s1(i);
    end loop;
    if d='1' then ax:=ap -(ax-ap); end if;
    return ax;
  end sf;

  procedure sbufprintf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                       file stream: text;    fmt: IN string;    s1: IN string) is
    variable debug:               boolean:=false;
    variable t1:                  string(1 to s1'length):=s1;
    variable d, f:                character;
    variable n, m, p:             integer;
    variable ai, i, j, ax, ap:    integer;
    variable fj:                  integer:=1;
    variable lflag, aflag, zflag: boolean;
    variable nflag, pflag, sflag: boolean;
    variable fmtflag:             boolean:=false;
    variable m1, m2, m3, m4:      string(1 to 64);
    variable t:                   string(1 to 128);
  begin
    if debug then
	write(buf, string'("-----------Starting sbufprintf Debug---------------"));
	writeline(output, buf);
    end if;
    loop
        fj:=fi;
	regmatch(ai, fi, fmt,
	"^$|\\n|([^%\\][^%\\]*)|%([^scdioxXufeEgGpn%\\]*)(.)|\\0(\\[0-7][0-7]?[0-7]?)|\\(.)", m1, m2);

	if debug then
	writeline(output, buf);
	write(buf, string'(" ai="));  write(buf, ai);
	write(buf, string'(" fi="));  write(buf, fi);
	write(buf, string'(" strlen(fmt)=")); write(buf, strlen(fmt));
	writeline(output, buf);
	write(buf, string'(" m1=")); write(buf, m1);
	write(buf, string'(" m2=")); write(buf, m2);
	writeline(output, buf);
	write(buf, string'("---------------------------------------------------"));
	writeline(output, buf);
	end if;

        case ai is --preprocess string
          when 5 => ai:=3;
            ax:=0;
            for i in 1 to m1'length loop
              ax:=ax*8;
              case m1(i) is
                when '0'=> ax:=ax+0; when '1'=> ax:=ax+1;
                when '2'=> ax:=ax+2; when '3'=> ax:=ax+3;
                when '4'=> ax:=ax+4; when '5'=> ax:=ax+5;
                when '6'=> ax:=ax+6; when '7'=> ax:=ax+7;
                when others =>
              end case;
            end loop;
            m1(1):=character'val(ax);

          when 6 => ai:=3;
            case m1(1) is
		when 'a'=> m1(1):=BEL; when 'b'=> m1(1):=BS;
		when 'f'=> m1(1):=FF;  when 'n'=> m1(1):=LF;
		when 'r'=> m1(1):=CR;  when 't'=> m1(1):=HT;
		when 'v'=> m1(1):=VT;  when others => 
            end case;
          when others =>
        end case;


	case ai is
	when 1 => exit;
	when 2 => if sbuf then write(buf, LF); else writeline(stream, buf); end if;
	when 3 => write(buf, m1(1 to strlen(m1)));
	when 4 =>
            if fmtflag then fi:=fj; exit; end if;
            f:=m2(1); m1(1):=NUL; m2(1):=NUL; m3(1):=NUL; m4(1):=NUL;
            regmatch(ai, fj, fmt,
                     "%([ 0#+-]*)([0-9]*)\.?([0-9]*).", m1, m2, m3, m4);

            --<m1:flags><m2:min print width>.<m3:precision:max strlen><modifier>
            lflag:=false; aflag:=false;
            nflag:=false; pflag:=false; sflag:=false; zflag:=false;


            for i in 1 to m1'length loop
              if m1(i)='-' then lflag:=true; end if;
              if m1(i)='+' then pflag:=true; end if;
              if m1(i)=' ' then sflag:=true; end if;
              if m1(i)='0' then zflag:=true; end if;
              if m1(i)='#' then aflag:=true; end if; --ignored in decimal
              if m1(i)=NUL then exit;        end if;
            end loop;
            if m2(1)=NUL then m:=0; else m:=atoi(m2); end if;

            if debug then
              writeline(output, buf);
              write(buf, string'(" min width="));  write(buf, m);
              write(buf, string'(" m1_flags=")); write(buf, m1);
              write(buf, string'(" m2_min_width=")); write(buf, m2);
              write(buf, string'(" m3_precision=")); write(buf, m3);
              writeline(output, buf);
            end if;

            if f='x' or f='o' then --unlimited string length
              if f='x' then ap:=4; else ap:=3; end if;

              ax:=0; n:=0; i:=0;
              for j in 1 to s1'length loop
                if s1(j)=NUL then exit; end if;
                if s1(j)=' ' then exit; end if;
                if s1(j)='1' then ax:=2*ax+1; else ax:=2*ax; end if; 
                n:=n+1;
                if n=ap then i:=i+1; m1(i):=pf_hex(ax); ax:=0; n:=0; end if;
              end loop;
              if n>0             then i:=i+1; m1(i):=pf_hex(ax); end if; --residue
              if f='x' and aflag then i:=i+1; m1(i):='x'; i:=i+1; m1(i):='0'; end if;
              if f='o' and aflag then i:=i+1; m1(i):='0'; end if;

              for j in 1 to i loop t1(j):=m1(i-j+1); end loop; --reverse string
              if i<t1'length then i:=i+1; t1(i):=NUL; end if;
              f:='s';

            elsif f='d' or f='u' then
              ax:=0; d:=' ';
              for i in 1 to s1'length loop
                if s1(i)=NUL then exit;                 end if;
                if s1(i)=' ' then exit;                 end if;
                if i=1       then ap:=1; else ap:=ap*2; end if;
                if s1(i)='1' then ax:=ax+ap;            end if;
                d:=s1(i);
              end loop;
              if f='d' and d='1' then nflag:=true; ax:=ap -(ax-ap); end if;

              i:=0;
              loop
                n:=ax mod 10;
                if i<m1'length then i:=i+1; m1(i):=pf_hex(n); end if;
                ax:=ax/10;
                if ax=0 then exit; end if;
              end loop;

              if zflag and not lflag then
                while i<m-1 loop i:=i+1; m1(i):='0'; end loop;
              end if; 
              if nflag                             then i:=i+1; m1(i):='-'; end if;
              if pflag and not nflag               then i:=i+1; m1(i):='+'; end if;
              if sflag and not pflag and not nflag then
                i:=i+1; if zflag then m1(i):='0'; else m1(i):=' '; end if;
              end if;
              if zflag and not lflag and i<m       then i:=i+1; m1(i):='0'; end if;

              for j in 1 to i loop t1(j):=m1(i-j+1); end loop;

              if i<t1'length then i:=i+1; t1(i):=NUL; end if;
              f:='s';
            end if;

            if f='%' then write(buf, '%'); end if;

            if f='s' then
              n:=strlen(t1);
              if m3(1)=NUL then p:=n; else p:=atoi(m3); end if;

	      if debug then
                writeline(output, buf);
	        write(buf, string'(" i=")); write(buf, n);
                write(buf, string'(" lflag=")); write(buf, lflag);
	        write(buf, string'(" min width="));  write(buf, m);
                write(buf, string'(" precision="));  write(buf, p);
                writeline(output, buf);
              end if;

              if not lflag AND p<m then
                write(buf, string'(" "), LEFT, m-p);
              end if;
                
              if n<=p then write(buf, t1(1 to n)); --no truncation
              else         write(buf, t1(1 to p)); end if;

              if lflag AND p<m then
                write(buf, string'(" "), LEFT, m-p);
              end if;
            end if;
            if f/='%' then fmtflag:=true; end if;
	when others =>
	end case;
    end loop;
    if debug then
	write(buf, string'("-----------Ending sbufprintf Debug-----------------"));
	writeline(output, buf);
    end if;
 end sbufprintf;

 procedure printf(format: IN string) is
   variable fi: integer:=1; variable a1: string(1 to 1);
 begin
   sbufprintf(false, fi, printf_buffer, output, format,  a1);
 end printf;

 procedure fprintf(sbuf:    boolean;
                   buf:     INOUT line;
                   file stream:  text;
	           format:  IN    string;  
	           a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : in string := " ";
	           a9,  a10, a11, a12, a13, a14, a15, a16: in string := " "
	   ) is
    variable fi: integer:=1;
  begin
    sbufprintf(sbuf, fi, buf, stream, format,  a1);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a2);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a3);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a4);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a5);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a6);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a7);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a8);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a9);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a10);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a11);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a12);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a13);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a14);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a15);
    if  fi<=format'length then sbufprintf(sbuf, fi, buf, stream, format,  a16);
    end if; end if; end if; end if; end if; end if; end if; end if;
    end if; end if; end if; end if; end if; end if; end if;
  end fprintf;

  procedure fprintf(buf:     INOUT line;  --concurrent version
                    file stream:  text;
	            format:  IN    string;  
	            a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : in string := " ";
	            a9,  a10, a11, a12, a13, a14, a15, a16: in string := " "
	     ) is
  begin
    fprintf(false, buf, stream, format, a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8,
                                      a9,  a10, a11, a12, a13, a14, a15, a16);
  end fprintf;

  procedure fprintf(file stream:  text; --limited to 1 output file at a time
	            format:  IN    string;  
	            a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : in string := " ";
	            a9,  a10, a11, a12, a13, a14, a15, a16: in string := " "
	     ) is
  begin
    fprintf(false, fprintf_buffer, stream, format, a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8,
                                      a9,  a10, a11, a12, a13, a14, a15, a16);
  end fprintf;

  procedure sprintf(s: INOUT line; format: IN string;  
	           a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : in string := " ";
	           a9,  a10, a11, a12, a13, a14, a15, a16: in string := " "
	   ) is
    variable fi: integer:=1;
  begin
    fprintf(true, s, output, format, a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8,
                                      a9,  a10, a11, a12, a13, a14, a15, a16);
  end sprintf;

  procedure sprintf(s: INOUT string; format: IN string;  
	            a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8 : in string := " ";
	            a9,  a10, a11, a12, a13, a14, a15, a16: in string := " "
	     ) is
    variable W: line; variable i: integer;
  begin
    fprintf(true, W, output, format, a1,  a2,  a3,  a4,  a5,  a6,  a7,  a8,
                               a9,  a10, a11, a12, a13, a14, a15, a16);
    i:=1; while i<=w'length loop
      if i<s'length then s(i):=w(i); i:=i+1; end if;
    end loop;
    if i<s'length then s(i):=NUL; end if;
    if w/=null then deallocate(W); end if;
  end sprintf;
 
  procedure sfgetc(c:    OUT   character; file stream: text;
                   buf:  INOUT line;      sbuf:   IN boolean) is
  begin
    if not sbuf then
      if buf=null then
        if not endfile(stream) then
          readline(stream, buf); write(buf, LF);
        end if;
      elsif buf'length<=0 then
        if not endfile(stream) then
          readline(stream, buf); write(buf, LF);
        end if;
      end if;
    end if;

    if buf=null then
       c:=NUL;
    else
      if buf'length<=0 then c:=NUL; else read(buf, c); end if;
    end if;
  end sfgetc;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT line) is
    variable debug:         boolean:=false;
    variable debugf:        line;
    variable c, d, f:       character;
    variable n, m, ms:      integer;
    variable ai, i, j, ax:  integer;
    variable fj:            integer:=1;
    variable sflag, mflag:  boolean;
    variable fmtflag:       boolean:=false;
    variable m1, m2, m3, m4:string(1 to 64);
  begin
    sfgetc(c, stream, buf, sbuf);
    loop
        fj:=fi; m1(1):=NUL; m2(1):=NUL;
	regmatch(ai, fi, fmt,
	"^$|%%|([^%\\])|%([^scdioxXufeEgGpn%\\]*)(.)|\\0(\\[0-7][0-7]?[0-7]?)|\\(.)", m1, m2);

	if debug then
	  writeline(output, debugf);
	  write(debugf, string'(" ai="));  write(debugf, ai);
	  write(debugf, string'(" fi="));  write(debugf, fi);
	  write(debugf, string'(" strlen(fmt)=")); write(debugf, strlen(fmt));
	  writeline(output, debugf);
	  write(debugf, string'(" m1=")); write(debugf, m1);
	  write(debugf, string'(" m2=")); write(debugf, m2);
	  writeline(output, debugf);
          write(debugf, string'(" buf="));  write(debugf, buf.ALL);
	  writeline(output, debugf);
	  write(debugf, string'("---------------------------------------------------"));
	  writeline(output, debugf);
	end if;

        case ai is
          when 2 => ai:=2; m1(1):='%'; m1(2):=NUL;

          when 5 => ai:=2;
            ax:=0;
            for i in 1 to m1'length loop
              ax:=ax*8;
              case m1(i) is
                when '0'=> ax:=ax+0; when '1'=> ax:=ax+1;
                when '2'=> ax:=ax+2; when '3'=> ax:=ax+3;
                when '4'=> ax:=ax+4; when '5'=> ax:=ax+5;
                when '6'=> ax:=ax+6; when '7'=> ax:=ax+7;
                when others =>
              end case;
            end loop;
            m1(1):=character'val(ax); m1(2):=NUL;

          when 6 => ai:=2;
            case m1(1) is
		when 'a'=> m1(1):=BEL; when 'b'=> m1(1):=BS;
		when 'f'=> m1(1):=FF;  when 'n'=> m1(1):=LF;
		when 'r'=> m1(1):=CR;  when 't'=> m1(1):=HT;
		when 'v'=> m1(1):=VT;  when others => 
            end case;
            m1(2):=NUL;
          when others =>
        end case;

	case ai is
	when 1 => exit;
	when 3 => 
	    if debug then
	      write(debugf, string'(" match c=")); write(debugf, c);
	      write(debugf, string'(" with m(1)=")); write(debugf, m1(1));
	      writeline(output, debugf);
            end if;
            if c=m1(1) then
              if c/=' ' then
                sfgetc(c, stream, buf, sbuf);
              else
                while c=' ' loop
                  sfgetc(c, stream, buf, sbuf);
	          if debug then
	            write(debugf, string'(" match loop c=")); write(debugf, c);
	            writeline(output, debugf);
                  end if;
                end loop;
              end if;
            else 
              exit;
            end if;
	when 4 =>
            if fmtflag then fi:=fj; exit; end if;

            f:=m2(1); m1(1):=NUL; m2(1):=NUL; m3(1):=NUL; m4(1):=NUL;
            regmatch(ai, fj, fmt, "%([*]*)([0-9]*)([lLh]).", m1, m2, m3, m4);

            --<m1:flags><m2:max scanf width><modifier><conversion type>
            sflag:=false;
            for i in 1 to m1'length loop
              if m1(i)='*' then sflag:=true; end if; --assignment suppression
              if m1(i)=NUL then exit;        end if;
            end loop;
            if m2(1)=NUL then m:=0; else m:=atoi(m2); end if;

            if debug then
              writeline(output, debugf);
              write(debugf, string'(" conversion=%"));  write(debugf, f);
              write(debugf, string'(" max width="));    write(debugf, m);
              write(debugf, string'(" m1_flags="));     write(debugf, m1);
              write(debugf, string'(" m2_max_width=")); write(debugf, m2);
              write(debugf, string'(" m3_modifier="));  write(debugf, m3);
              writeline(output, debugf);
            end if;

            if f/='c' then
              while c=' ' OR c=LF OR c=HT OR c=CR OR c=VT loop
                sfgetc(c, stream, buf, sbuf);
              end loop;
            end if;

            if f='x' OR f='o' OR f='d' then 
              if f='x' or f='o' then write(s1, string'("0")); end if;

              i:=0; ax:=0; ms:=1;
              loop
--                write(debugf, string'(" c="));  write(debugf, c);
--                write(debugf, string'(" i="));  write(debugf, i);
--                write(debugf, string'(" m="));  write(debugf, m);
--                write(debugf, string'(" buf="));  write(debugf, buf.ALL);
--                writeline(output, debugf);
                if m/=0 AND i>=m then exit; end if;
                if    i=0 and c='+' then ms:=1;
                elsif i=0 and c='-' then ms:=-1;
                else
                  case c is
                    when '0'    => n:=0;  when '1'    => n:=1;
                    when '2'    => n:=2;  when '3'    => n:=3;
                    when '4'    => n:=4;  when '5'    => n:=5;
                    when '6'    => n:=6;  when '7'    => n:=7;
                    when '8'    => n:=8;  when '9'    => n:=9;
                    when 'a'|'A'=> n:=10; when 'b'|'B'=> n:=11;
                    when 'c'|'C'=> n:=12; when 'd'|'D'=> n:=13;
                    when 'e'|'E'=> n:=14; when 'f'|'F'=> n:=15;
                    when others => exit;
                  end case;
                  if f='d' then
                    if n>9 then exit; else ax:=ax*10+n; end if;
                  else 
                    m1(1 to 4):=pf_hex2(n);
                    if f='o' and n>7 then exit; end if;
                    if f='o' and not sflag then write(s1, m1(1 to 3)); end if;
                    if f='x' and not sflag then write(s1, m1(1 to 4)); end if;
                  end if;
                end if;
                i:=i+1; sfgetc(c, stream, buf, sbuf);
              end loop;

              if (f='o' or f='x') and not sflag then
                for i in 1 to s1'length/2 loop
                  d:=s1(i); s1(i):=s1(s1'length+1-i); s1(s1'length+1-i):=d;
                end loop;
              end if;

              if (f='d' or f='u') then	-- MIKE did this and the 'u' is wrong!
                m1:=pf(ms*ax); n:=strlen(m1); write(s1, m1(1 to n));
              end if;
            end if;

            if f='%' then write(buf, '%'); end if;

            if f='s' then
              i:=0;
              while c/=' ' AND c/=LF AND c/=HT AND c/=CR AND c/=VT AND c/=NUL loop
                if m/=0 AND i>=m then exit; end if;
                if not sflag then write(s1, c); end if;
                i:=i+1; sfgetc(c, stream, buf, sbuf);
              end loop;

	      if debug then
                writeline(output, debugf);
                write(debugf, string'(" sflag=")); write(debugf, sflag);
                write(debugf, string'(" s1=")); write(debugf, s1.all);
                writeline(output, debugf);
              end if;
            end if;
            if c/='%' AND not sflag then fmtflag:=true; end if;
	when others =>
	end case;
    end loop;
    write(debugf, c & buf.all);
    if buf/=null then deallocate(buf); end if;
    write(buf, debugf.all); deallocate(debugf);
  end sbufscanf;

  procedure sbufscanf_id(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT integer) is
    variable w: line;
  begin
    s1 := 0;
    sbufscanf(sbuf, fi, buf, stream, fmt, w);
    if w/=null then if w'length>0 then s1:=1; end if; end if;
    deallocate(w); 
  end sbufscanf_id;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT integer) is
    variable w: line;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then if w'length>0 then s1:=sf(w.all); end if; end if;
    deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT character) is
    variable w: line;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then if w'length>0 then s1:=w(1); end if; end if;
    deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT string) is
    variable w: line; variable i: integer:=1;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then
      while i<=w'length loop
        if i<s1'length then s1(i):=w(i); i:=i+1; end if;
      end loop;
    end if;
    if i<s1'length then s1(i):=NUL; end if; deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT bit) is
    variable w: line;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then 
      if w'length>0 then if w(1)='1' then s1:='1'; else s1:='0'; end if; end if;
    end if;
    deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT boolean) is
    variable w: line;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then 
      if w'length>0 then if w(1)='1' then s1:=true; else s1:=false; end if; end if;
    end if;
    deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT bit_vector) is
    variable w: line; variable i, j: integer;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then
      j:=s1'left; for i in 1 to w'length loop
        if s1'low<=j AND j<=s1'high then          
          if w(i)='1' then s1(j):='1'; else s1(j):='0'; end if;
          if s1'left<=s1'right then j:=j+1; else j:=j-1; end if;
        else
          exit;
        end if;
      end loop;
    end if;
    deallocate(W); 
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT std_logic) is
    variable w: line;
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);
    if w/=null then
      if w'length>0 then 
        case w(1) is
          when '1'    => s1:='1';
          when '0'    => s1:='0';
          when '-'    => s1:='-';
          when 'X'|'x'=> s1:='X';
          when 'Z'|'z'=> s1:='Z';
          when 'H'|'h'=> s1:='H';
          when 'L'|'l'=> s1:='L';
          when 'W'|'w'=> s1:='W';
          when others => s1:='U';
        end case;
      end if;
    end if;
    deallocate(w);
  end sbufscanf;

  procedure sbufscanf(sbuf:   IN    boolean; fi: INOUT integer; buf: INOUT line;
                      file stream: text;    fmt: IN string;    s1:  INOUT std_logic_vector) is
    variable w: line; variable i, j: integer;
    variable l : line;
    variable c : character;
    variable last : character := 'U';
  begin
    sbufscanf(sbuf, fi, buf, stream, fmt, W);

    if (w = NULL) then
      for i in s1'range loop
        s1(i) := 'U';
      end loop;
    else

--      write(l,string'("sbufscanf(std_logic_vector): "));
--      write(l,string'(" s1'left=")); write(l,s1'left);
--      write(l,string'(" s1'right=")); write(l,s1'right);
--      write(l,string'(" s1'length=")); write(l,s1'length);
--      write(l,string'(" fi=")); write(l,fi);
--      write(l,string'(" fmt=")); write(l,fmt);
--      write(l,string'(" w=")); write(l,w.ALL);
--      writeline(output, l);
    
      if ( s1'left <= s1'right ) then
        j:=s1'left;
      else
        j:=s1'right;
      end if;
      i := 1;
      loop
        if i <= w'length then c := w(i); else c := last; end if;
        if s1'low<=j AND j<=s1'high then 
--          write(l,string'(" i=")); write(l,i);
--          write(l,string'(" j=")); write(l,j);
--          write(l,string'(" wi=")); write(l,c);
--          writeline(output, l);
          case c is
            when '1'    => s1(j):='1';
            when '0'    => s1(j):='0';
            when '-'    => s1(j):='-';
            when 'X'|'x'=> s1(j):='X';
            when 'Z'|'z'=> s1(j):='Z';
            when 'H'|'h'=> s1(j):='H';
            when 'L'|'l'=> s1(j):='L';
            when 'W'|'w'=> s1(j):='W';
            when others => s1(j):='U';
          end case;
          last := c;
          j:=j+1;
        else
          exit;
        end if;
        i := i + 1;
      end loop;

      deallocate(w);
    end if;
--    write(l,string'(" s1(0)=")); write(l,s1(0));
--    write(l,string'(" s1(1)=")); write(l,s1(1));
--    write(l,string'(" s1(2)=")); write(l,s1(2));
--    write(l,string'(" s1=")); write(l,s1);
--    writeline(output, l);
  end sbufscanf;

  procedure fclose(buf: INOUT line; file stream: text) is
  begin
    if buf/=null then
      if buf'length > 0 then
        writeline(stream, buf);
      end if;
    end if;
    FILE_CLOSE(stream);
  end fclose;

  procedure fclose(file stream: text) is
  begin
    fclose(fprintf_buffer, stream);
  end fclose;

  procedure fgetc( c: OUT character; file stream: text; buf: INOUT line) is
  begin
    sfgetc(c, stream, buf, false);
  end fgetc;

  procedure fgetc( c: OUT character; file stream: text) is
  begin
    sfgetc(c, stream, fscanf_buffer, false);
  end fgetc;

  procedure getc(  c: OUT character; file stream: text) is
  begin
    sfgetc(c, stream, fscanf_buffer, false);
  end getc;

  procedure ungetc(c: IN character; file stream: text; buf: INOUT line) is
    variable temp: line;
  begin
    write(temp, c & buf.all);
    if buf/=null then deallocate(buf); end if;
    write(buf, temp.all); deallocate(temp);
  end ungetc;

  procedure ungetc(c: IN character; file stream: text) is
    variable temp: line;
  begin
    ungetc(c, stream, fscanf_buffer);
  end ungetc;

  procedure getchar(c: OUT character) is
  begin
    sfgetc(c, input, fscanf_buffer, false);
  end getchar;

  procedure fgets(s: OUT  string; n: IN integer; file stream: text; buf: INOUT line) is
    variable i: integer:=0; variable c: character;
  begin
    while i<=n loop
      if i>=s'length then exit; end if;
      sfgetc(c, stream, buf, false);
      i:=i+1; s(i):=c; if c=LF OR c=NUL then exit; end if;
    end loop; 
    if i<s'length then i:=i+1; s(i):=NUL; end if;
  end fgets;

  procedure fgets(s: OUT  string; n: IN integer; file stream: text) is
  begin
    fgets(s, n, stream, fscanf_buffer);
  end fgets;
 
  procedure gets(s: OUT string) is --used for console or stdin
    variable buf: line;
  begin
    readline(input, buf); strcpy(s, buf.all); deallocate(buf);
  end gets;

  procedure fputc(c: IN character; file stream: text; buf: INOUT line) is
  begin
    if c=LF then
      if buf/=null then
        if buf'length>0 then writeline(stream, buf); end if;
      end if;
    else
      write(buf, c);
    end if;
  end fputc;

  procedure fputc(c: IN character; file stream: text) is
  begin
    fputc(c, stream, fprintf_buffer);
  end fputc;

  procedure putc(c: IN character; file stream: text) is
  begin
    fputc(c, stream);
  end putc;

  procedure putchar(c: IN character) is
  begin
     fputc(c, output, printf_buffer);
  end putchar;

  procedure fputs(s: IN string; file stream: text; buf: INOUT line) is
    variable i, n: integer;
  begin
    n:=strlen(s);
    for i in 1 to n loop fputc(s(i), stream, buf); end loop;
  end fputs;

  procedure fputs(s: IN string; file stream: text) is
  begin
    fputs(s, stream, fprintf_buffer);
  end fputs;

  procedure puts(s: IN string) is
  begin
    write(printf_buffer, s(1 to strlen(s))); writeline(output, printf_buffer);
  end puts;

  procedure printf(format: IN string; a1: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
  end printf;

  procedure printf(format: IN string; a1: string; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: string) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: integer) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: std_logic) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: boolean) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: bit) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: string; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: integer; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: std_logic_vector; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: boolean; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure printf(format: IN string; a1: bit_vector; a2: bit_vector) is
    variable fi: integer:=1;
  begin
    sbufprintf(false, fi, printf_buffer, output, format, pf(a1));
    sbufprintf(false, fi, printf_buffer, output, format, pf(a2));
  end printf;

  procedure scanf(format: IN string; a1: INOUT string) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT integer) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT std_logic) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT boolean) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT bit) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure scanf(format: IN string; a1: INOUT bit_vector) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, scanf_buffer, input, format, a1);
  end scanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT string) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT integer) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT std_logic) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT std_logic_vector) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT boolean) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT bit) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  procedure fscanf(file stream: text; format: IN string; a1: INOUT bit_vector) is
    variable fi: integer:=1;
  begin
    sbufscanf(false, fi, fscanf_buffer, stream, format, a1);
  end fscanf;

  impure function sscanf(s: string; format: string) return integer is
    variable W: line; variable fi: integer:=1;
    variable a1: integer := 0;
  begin
    write(W, s);
    sbufscanf_id(true, fi, W, output, format, a1);
    deallocate(W);
    return a1;
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT string) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT integer) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT std_logic) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT std_logic_vector) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT boolean) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT bit) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

  procedure sscanf(s: IN string; format: IN string; a1: INOUT bit_vector) is
    variable W: line; variable fi: integer:=1;
  begin
    write(W, s);
    sbufscanf(true, fi, W, output, format, a1);
    deallocate(W);
  end sscanf;

end stdio_h;
